"""
Detect jumps in a science image
"""

import logging
import time

import numpy as np
from roman_datamodels import datamodels as rdd
from stcal.jump.jump import detect_jumps

from romancal.lib import dqflags
from romancal.stpipe import RomanStep

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


__all__ = ["JumpStep"]


class JumpStep(RomanStep):
    """
    JumpStep: Performs CR/jump detection. The 2-point difference method is
    applied.
    """

    spec = """
        rejection_threshold = float(default=180.0,min=0) # CR sigma rej thresh
        three_group_rejection_threshold = float(default=185.0,min=0) # CR sigma rej thresh
        four_group_rejection_threshold = float(default=190.0,min=0) # CR sigma rej thresh
        maximum_cores = option('none', 'quarter', 'half', 'all', default='none') # max number of processes to create
        flag_4_neighbors = boolean(default=True) # flag the four perpendicular neighbors of each CR
        max_jump_to_flag_neighbors = float(default=1000) # maximum jump sigma that will trigger neighbor flagging
        min_jump_to_flag_neighbors = float(default=10) # minimum jump sigma that will trigger neighbor flagging
        min_sat_area = float(default=1.0) # minimum required area for the central saturation of snowballs
        min_jump_area = float(default=5.0) # minimum area to trigger large events processing
        expand_factor = float(default=2.0) # The expansion factor for the enclosing circles or ellipses
        use_ellipses = boolean(default=False) # Use an enclosing ellipse rather than a circle for MIRI showers
        sat_required_snowball = boolean(default=True) # Require the center of snowballs to be saturated
        expand_large_events = boolean(default=False) # must be True to trigger snowball and shower flagging
        use_ramp_jump_detection = boolean(default=True) # Use jump detection during ramp fitting
    """  # noqa: E501

    reference_file_types = ["gain", "readnoise"]

    def process(self, input):
        # Open input as a Roman DataModel (single integration; 3D arrays)

        with rdd.open(input, lazy_load=False) as input_model:
            # Extract the needed info from the Roman Data Model
            meta = input_model.meta
            r_data = input_model.data.value
            r_gdq = input_model.groupdq
            r_pdq = input_model.pixeldq
            r_err = input_model.err.value
            result = input_model

            # If the ramp fitting jump detection is enabled, then skip this step
            if self.use_ramp_jump_detection:
                result.meta.cal_step.jump = "SKIPPED"
                return result

            frames_per_group = meta.exposure.nframes

            # Modify the arrays for input into the 'common' jump (4D)
            data = np.copy(r_data[np.newaxis, :])
            gdq = r_gdq[np.newaxis, :]
            pdq = r_pdq[np.newaxis, :]
            err = np.copy(r_err[np.newaxis, :])

            tstart = time.time()

            # Check for an input model with NGROUPS<=2
            ngroups = data.shape[1]

            if ngroups <= 2:
                self.log.warning("Cannot apply jump detection as NGROUPS<=2;")
                self.log.warning("Jump step will be skipped")

                result = input_model

                result.meta.cal_step.jump = "SKIPPED"
                return result

            # Retrieve the parameter values
            rej_thresh = self.rejection_threshold
            three_grp_rej_thresh = self.three_group_rejection_threshold
            four_grp_rej_thresh = self.four_group_rejection_threshold
            max_cores = self.maximum_cores
            max_jump_to_flag_neighbors = self.max_jump_to_flag_neighbors
            min_jump_to_flag_neighbors = self.min_jump_to_flag_neighbors
            flag_4_neighbors = self.flag_4_neighbors
            min_sat_area = self.min_sat_area
            min_jump_area = self.min_jump_area
            expand_factor = self.expand_factor
            use_ellipses = self.use_ellipses
            sat_required_snowball = self.sat_required_snowball
            expand_large_events = self.expand_large_events

            self.log.info("CR rejection threshold = %g sigma", rej_thresh)
            if self.maximum_cores != "none":
                self.log.info("Maximum cores to use = %s", max_cores)

            # Get the gain and readnoise reference files
            gain_filename = self.get_reference_file(input_model, "gain")
            self.log.info("Using GAIN reference file: %s", gain_filename)
            gain_model = rdd.GainRefModel(gain_filename)
            gain_2d = gain_model.data.value

            readnoise_filename = self.get_reference_file(input_model, "readnoise")
            self.log.info("Using READNOISE reference file: %s", readnoise_filename)
            readnoise_model = rdd.ReadnoiseRefModel(readnoise_filename)
            # This is to clear the WRITEABLE=False flag?
            readnoise_2d = np.copy(readnoise_model.data.value)

            # DG 0810/21:  leave for now; make dqflags changes in a later,
            #              separate PR
            dqflags_d = {}  # Dict of DQ flags
            dqflags_d = {
                "GOOD": dqflags.group["GOOD"],
                "DO_NOT_USE": dqflags.group["DO_NOT_USE"],
                "SATURATED": dqflags.group["SATURATED"],
                "JUMP_DET": dqflags.group["JUMP_DET"],
                "NO_GAIN_VALUE": dqflags.pixel["NO_GAIN_VALUE"],
            }

            gdq, pdq, *_ = detect_jumps(
                frames_per_group,
                data,
                gdq,
                pdq,
                err,
                gain_2d,
                readnoise_2d,
                rej_thresh,
                three_grp_rej_thresh,
                four_grp_rej_thresh,
                max_cores,
                max_jump_to_flag_neighbors,
                min_jump_to_flag_neighbors,
                flag_4_neighbors,
                dqflags_d,
                min_sat_area=min_sat_area,
                min_jump_area=min_jump_area,
                expand_factor=expand_factor,
                use_ellipses=use_ellipses,
                sat_required_snowball=sat_required_snowball,
                expand_large_events=expand_large_events,
            )

            gdq = gdq[0, :, :, :]
            pdq = pdq[0, :, :]
            result.groupdq = gdq
            result.pixeldq = pdq
            gain_model.close()
            readnoise_model.close()
            tstop = time.time()
            self.log.info("The execution time in seconds: %f", tstop - tstart)

        result.meta.cal_step.jump = "COMPLETE"

        if self.save_results:
            try:
                self.suffix = "jump"
            except AttributeError:
                self["suffix"] = "jump"

        return result



#!/usr/bin/env python
import logging
from os.path import basename

import romancal.datamodels.filetype as filetype
from romancal.outlier_detection import OutlierDetectionStep
from romancal.resample import ResampleStep

# step imports
from romancal.skymatch import SkyMatchStep

from ..stpipe import RomanPipeline

__all__ = ["HighLevelPipeline"]

# Define logging
log = logging.getLogger()
log.setLevel(logging.DEBUG)


class HighLevelPipeline(RomanPipeline):
    """
    HighLevelPipeline: Apply all calibration steps to the roman data
    to produce level 3 products. Included steps are:
    ``skymatch``, ``outlier_detection`` and ``resample``.
    """

    class_alias = "roman_hlp"

    spec = """
        save_results = boolean(default=False)
    """

    # Define aliases to steps
    step_defs = {
        "skymatch": SkyMatchStep,
        "outlier_detection": OutlierDetectionStep,
        "resample": ResampleStep,
    }

    # start the actual processing
    def process(self, input):
        """Process the Roman WFI data from Level 2 to Level 3"""

        log.info("Starting Roman high level calibration pipeline ...")
        if isinstance(input, str):
            input_filename = basename(input)
        else:
            input_filename = None

        # open the input file
        file_type = filetype.check(input)
        if file_type == "asdf":
            log.info("The level three pipeline input needs to be an association")
            return

        if file_type == "asn":
            self.skymatch.suffix = "skymatch"
            result = self.skymatch(input)
            self.skymatch.suffix = "outlier_detection"
            # result = self.outlier_detection(input)
            self.skymatch.suffix = "i2d"
            result = self.resample(input)
            if input_filename:
                result.meta.filename = input_filename

        return result



import pytest
from metrics_logger.decorators import metrics_logger
from roman_datamodels import datamodels as rdm

from romancal.stpipe import RomanStep
from romancal.tweakreg.tweakreg_step import TweakRegStep

from .regtestdata import compare_asdf


@metrics_logger("DMS280")
@pytest.mark.bigdata
def test_tweakreg(rtdata, ignore_asdf_paths, tmp_path):
    # N.B.: the data were created using WFIsim and processed through
    # the three pipeline steps listed below:
    # - assign_wcs;
    # - photom;
    # - source_detection.
    input_data = "r0000101001001001001_01101_0001_WFI01_cal.asdf"
    output_data = "r0000101001001001001_01101_0001_WFI01_tweakregstep.asdf"
    truth_data = "r0000101001001001001_01101_0001_WFI01_tweakregstep.asdf"

    asn_fn = "tweakreg_asn.json"

    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.get_data(f"WFI/image/{asn_fn}")
    rtdata.get_truth(f"truth/WFI/image/{truth_data}")

    rtdata.input = asn_fn
    rtdata.output = output_data

    # instantiate TweakRegStep (for running and log access)
    step = TweakRegStep()

    args = [
        "romancal.step.TweakRegStep",
        rtdata.input,
        f"--output_file='{rtdata.output}'",
        "--suffix='tweakregstep'",
    ]
    RomanStep.from_cmdline(args)
    tweakreg_out = rdm.open(rtdata.output)

    step.log.info(
        "DMS280 MSG: TweakReg step recorded as complete? :"
        f' {tweakreg_out.meta.cal_step.tweakreg == "COMPLETE"}'
    )
    assert tweakreg_out.meta.cal_step.tweakreg == "COMPLETE"

    step.log.info(
        f"""DMS280 MSG: TweakReg created new attribute with fit results? :\
            {"wcs_fit_results" in tweakreg_out.meta}"""
    )
    assert "wcs_fit_results" in tweakreg_out.meta

    step.log.info(
        f"""DMS280 MSG: TweakReg created new coordinate frame 'v2v3corr'? :\
            {"v2v3corr" in tweakreg_out.meta.wcs.available_frames}"""
    )
    assert "v2v3corr" in tweakreg_out.meta.wcs.available_frames

    diff = compare_asdf(rtdata.output, rtdata.truth, atol=1e-3, **ignore_asdf_paths)
    step.log.info(
        f"DMS280 MSG: Was the proper TweakReg data produced? : {diff.identical}"
    )
    assert diff.identical, diff.report()



import pytest

from romancal.datamodels import ModelContainer
from romancal.outlier_detection import OutlierDetectionStep, outlier_detection


@pytest.mark.parametrize(
    "input_models",
    [
        list(),
        [""],
        "",
        "testing",
        {},
    ],
)
def test_outlier_raises_error_on_invalid_input_models(input_models):
    """Test that OutlierDetection raises an error if input is not a ModelContainer."""

    step = OutlierDetectionStep()

    step.process(input_models)

    assert step.skip is True


def test_outlier_skips_step_on_invalid_number_of_elements_in_input(base_image):
    img = base_image()

    step = OutlierDetectionStep()
    step.process(ModelContainer([img]))

    assert step.skip is True
    assert step.input_models[0].meta.cal_step.outlier_detection == "SKIPPED"


def test_outlier_skips_step_on_exposure_type_different_from_wfi_image(base_image):
    img_1 = base_image()
    img_1.meta.exposure.type = "WFI_PRISM"
    img_2 = base_image()
    img_2.meta.exposure.type = "WFI_PRISM"

    step = OutlierDetectionStep()
    step.process(ModelContainer([img_1, img_2]))

    assert step.input_models[0].meta.cal_step.outlier_detection == "SKIPPED"
    assert step.skip


def test_outlier_valid_input_datamodels(tmp_path, base_image):
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    img_2 = base_image()
    img_2.meta.filename = "img_2.asdf"

    step = OutlierDetectionStep()
    # set output dir for all files created by the step
    step.output_dir = tmp_path
    # make sure resample does not save file to disk
    step.in_memory = True
    step.process(ModelContainer([img_1, img_2]))

    assert step.skip is False
    assert img_1.meta.cal_step.outlier_detection == "COMPLETE"
    assert img_2.meta.cal_step.outlier_detection == "COMPLETE"


def test_outlier_valid_input_asn(tmp_path, base_image, create_mock_asn_file):
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    img_1.save(tmp_path / "img_1.asdf")
    img_2 = base_image()
    img_1.meta.filename = "img_2.asdf"
    img_2.save(tmp_path / "img_2.asdf")

    asn_filepath = create_mock_asn_file(tmp_path)

    step = OutlierDetectionStep()
    # set output dir for all files created by the step
    step.output_dir = tmp_path
    # make sure resample does not save file to disk
    step.in_memory = True
    step.process(ModelContainer(asn_filepath))

    assert step.skip is False
    assert all(
        x.meta.cal_step.outlier_detection == "COMPLETE" for x in step.input_models
    )


@pytest.mark.parametrize(
    "pars",
    [
        {
            "weight_type": "exptime",
            "pixfrac": 1.0,
            "kernel": "square",
            "fillval": "INDEF",
            "nlow": 0,
            "nhigh": 0,
            "maskpt": 0.7,
            "grow": 1,
            "snr": "4.0 3.0",
            "scale": "0.5 0.4",
            "backg": 0.0,
            "kernel_size": "7 7",
            "save_intermediate_results": False,
            "resample_data": True,
            "good_bits": 0,
            "allowed_memory": None,
            "in_memory": True,
            "make_output_path": None,
            "resample_suffix": "i2d",
        },
        {
            "weight_type": "exptime",
            "save_intermediate_results": True,
            "make_output_path": None,
            "resample_suffix": "some_other_suffix",
        },
    ],
)
def test_outlier_init_default_parameters(pars, base_image):
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    input_models = ModelContainer([img_1])

    step = outlier_detection.OutlierDetection(input_models, **pars)

    assert step.input_models == input_models
    assert step.outlierpars == pars
    assert step.make_output_path == pars["make_output_path"]
    assert step.resample_suffix == f"_outlier_{pars['resample_suffix']}.asdf"


def test_outlier_do_detection(tmp_path, base_image):
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    img_2 = base_image()
    img_2.meta.filename = "img_2.asdf"
    input_models = ModelContainer([img_1, img_2])

    outlier_step = OutlierDetectionStep()
    # set output dir for all files created by the step
    outlier_step.output_dir = tmp_path
    # make sure resample does not save file to disk
    outlier_step.in_memory = True

    pars = {
        "weight_type": "exptime",
        "pixfrac": 1.0,
        "kernel": "square",
        "fillval": "INDEF",
        "nlow": 0,
        "nhigh": 0,
        "maskpt": 0.7,
        "grow": 1,
        "snr": "4.0 3.0",
        "scale": "0.5 0.4",
        "backg": 0.0,
        "kernel_size": "7 7",
        "save_intermediate_results": False,
        "resample_data": True,
        "good_bits": 0,
        "allowed_memory": None,
        "in_memory": outlier_step.in_memory,
        "make_output_path": outlier_step.make_output_path,
        "resample_suffix": "i2d",
    }

    blot_path_1 = tmp_path / img_1.meta.filename.replace(".asdf", "_blot.asdf")
    blot_path_2 = tmp_path / img_2.meta.filename.replace(".asdf", "_blot.asdf")
    median_path = tmp_path / "img_median.asdf"

    outlier_files_path = [
        blot_path_1,
        blot_path_2,
        median_path,
    ]

    detection_step = outlier_detection.OutlierDetection
    step = detection_step(input_models, **pars)

    step.do_detection()

    assert all(x.exists() for x in outlier_files_path)
